const nugget = require('nugget')
const assert = require('assert')
const fs = require('fs')
const tar = require('tar')
const co = require('co')
const npminstall = require('npminstall')
const { join } = require('path')
const { arch, platform } = require('os')

const exists = file => new Promise((resolve, reject) =>
  fs.access(file, err => err ? reject(err) : resolve(file)))
const toJson = file => new Promise((resolve, reject) =>
  fs.readFile(file, (err, data) => err ? reject(err) : resolve(JSON.parse(data))))

module.exports = (opts = {}) => {
  ['name', 'version'].forEach(name => assert(opts[name], `the opts.${name} is required`))
  opts = Object.assign({
    root: join(process.cwd(), 'node_modules', opts.name),
    cwd: process.cwd(),
    dir: `${__dirname}/cache`,
    arch: arch(),
    platform: platform(),
    modules: process.versions.modules,
    mirror: process.env.NPM_CONFIG_BIN_BASE ||
      process.env.npm_config_bin_base ||
      process.env.BIN_BASE ||
      opts.mirror ||
      'https://bdf.mengb.net/',
    registry: process.env.NPM_CONFIG_REGISTRY ||
      process.env.npm_config_registry ||
      process.env.REGISTRY ||
      opts.registry
  }, opts)

  let filename = `node-v${opts.modules}-${opts.platform}-${opts.arch}.tar.gz`
  let target = `${opts.name}@${opts.version}-${filename}`
  let file = join(opts.dir, target)
  let bin = join(opts.root, '.bin-downloader')
  let jsonInfo = JSON.stringify({
    version: opts.version,
    arch: opts.arch,
    platform: opts.platform,
    time: new Date()
  }, null, 2)

  return co(function * () {
    try {
      yield exists(bin)
      let json = yield toJson(bin)
      assert(
        json.version === opts.version &&
        json.arch === opts.arch &&
        json.platform === opts.platform
      )
    } catch (e) {
      try {
        yield exists(opts.dir)
      } catch (e1) {
        yield new Promise((resolve, reject) =>
          fs.mkdir(opts.dir, err => err ? reject(err) : resolve()))
      }
      try {
        yield exists(file)
      } catch (e1) {
        try {
          yield new Promise((resolve, reject) => nugget(
            `${opts.mirror.replace(/\/$/, '')}/${opts.name}/v${opts.version}/${filename}`,
            Object.assign(opts, { target, resume: true }),
            err => err ? reject(err) : resolve()
          ))
        } catch (e2) {
          yield npminstall({
            registry: opts.registry,
            root: opts.cwd,
            pkgs: [{ name: opts.name, version: opts.version }]
          })
          yield new Promise((resolve, reject) => fs.writeFile(
            bin,
            jsonInfo,
            err => err ? reject(err) : resolve()))
          return
        }
      }
      try {
        let file = join(opts.root, 'package.json')
        yield exists(file)
        assert((yield toJson(file)).version === opts.version)
      } catch (e2) {
        yield npminstall({
          ignoreScripts: true,
          registry: opts.registry,
          root: opts.cwd,
          pkgs: [{ name: opts.name, version: opts.version }]
        })
      }
      yield tar.x({ file, cwd: opts.root })
      yield new Promise((resolve, reject) => fs.writeFile(
        bin,
        jsonInfo,
        err => err ? reject(err) : resolve()))
    }
  })
}
